# License: MIT
# Copyright © 2022 Frequenz Energy-as-a-Service GmbH

"""Generate the code reference pages.

It uses the following `mkdocs` plugins:

* `mkdocs-gen-files` to generate the API documentation pages.
* `mkdocs-literate-nav` to make use of the generate `SUMMARY.md` file.

Based on the recipe at:
https://mkdocstrings.github.io/recipes/#automatic-code-reference-pages
"""

import subprocess
import tempfile
from pathlib import Path
from typing import Tuple

import mkdocs_gen_files

from . import protobuf as _protobuf


def _is_internal(path_parts: Tuple[str, ...]) -> bool:
    """Tell if the path is internal judging by the parts.

    Args:
        path_parts: Path.parts of the path to check.

    Returns:
        True if the path is internal.
    """

    def with_underscore_not_init(part: str) -> bool:
        return part.startswith("_") and part != "__init__"

    return any(p for p in path_parts if with_underscore_not_init(p))


def generate_python_api_pages(
    src_path: str = "src", dst_path: str = "python-reference"
) -> None:
    """Generate API documentation pages for the code.

    Internal modules (those starting with an underscore except from `__init__`) are
    not included.

    A summary page is generated as `SUMMARY.md` which is compatible with the
    `mkdocs-literary-nav` plugin.

    Args:
        src_path: Path where the code is located.
        dst_path: Path where the documentation should be generated.  This is relative
            to the output directory of mkdocs.
    """
    # type ignore because mkdocs_gen_files uses a very weird module-level
    # __getattr__() which messes up the type system
    nav = mkdocs_gen_files.Nav()  # type: ignore

    for path in sorted(Path(src_path).rglob("*.py")):
        module_path = path.relative_to(src_path).with_suffix("")

        doc_path = path.relative_to(src_path).with_suffix(".md")
        full_doc_path = Path(dst_path, doc_path)
        parts = tuple(module_path.parts)
        if _is_internal(parts):
            continue
        if parts[-1] == "__init__":
            doc_path = doc_path.with_name("index.md")
            full_doc_path = full_doc_path.with_name("index.md")
            parts = parts[:-1]

        nav[parts] = doc_path.as_posix()

        with mkdocs_gen_files.open(full_doc_path, "w") as output_file:
            output_file.write(f"::: {'.'.join(parts)}\n")

        mkdocs_gen_files.set_edit_path(full_doc_path, Path("..") / path)

    with mkdocs_gen_files.open(Path(dst_path) / "SUMMARY.md", "w") as nav_file:
        nav_file.writelines(nav.build_literate_nav())


def generate_protobuf_api_pages(
    src_path: str = "proto", dst_path: str = "protobuf-reference"
) -> None:
    """Generate API documentation pages for the code.

    Internal modules (those starting with an underscore except from `__init__`) are
    not included.

    A summary page is generated as `SUMMARY.md` which is compatible with the
    `mkdocs-literary-nav` plugin.

    Args:
        src_path: Path where the code is located.
        dst_path: Path where the documentation should be generated.  This is relative
            to the output directory of mkdocs.
    """
    # type ignore because mkdocs_gen_files uses a very weird module-level
    # __getattr__() which messes up the type system
    nav = mkdocs_gen_files.Nav()  # type: ignore
    config = _protobuf.ProtobufConfig.from_pyproject_toml(
        proto_path=src_path, docs_path=dst_path
    )

    cwd = Path.cwd()

    with tempfile.TemporaryDirectory(prefix="mkdocs-protobuf-reference-") as tmp_path:
        for path in sorted(Path(config.proto_path).rglob("*.proto")):
            doc_path = path.relative_to(config.proto_path).with_suffix(".md")
            full_doc_path = Path(config.docs_path, doc_path)
            parts = tuple(path.relative_to(config.proto_path).parts)
            nav[parts] = doc_path.as_posix()
            doc_tmp_path = tmp_path / doc_path
            doc_tmp_path.parent.mkdir(parents=True, exist_ok=True)
            try:
                subprocess.run(
                    [
                        "docker",
                        "run",
                        "--rm",
                        f"-v{cwd}:{cwd}",
                        f"-v{tmp_path}:{tmp_path}",
                        "pseudomuto/protoc-gen-doc",
                        f"-I{cwd / config.proto_path}",
                        *(f"-I{cwd / p}" for p in config.include_paths),
                        f"--doc_opt=markdown,{doc_path.name}",
                        f"--doc_out={tmp_path / doc_path.parent}",
                        str(cwd / path),
                    ],
                    check=True,
                )
            except subprocess.CalledProcessError as error:
                print(f"Error generating protobuf reference page: {error}")

            with doc_tmp_path.open() as input_file, mkdocs_gen_files.open(
                full_doc_path, "w"
            ) as output_file:
                output_file.write(input_file.read())

            mkdocs_gen_files.set_edit_path(full_doc_path, Path("..") / path)

    with mkdocs_gen_files.open(Path(config.docs_path) / "SUMMARY.md", "w") as nav_file:
        nav_file.writelines(nav.build_literate_nav())



# License: MIT
# Copyright © 2023 Frequenz Energy-as-a-Service GmbH

"""Manages the configuration to generate files from the protobuf files."""

import dataclasses
import logging
import pathlib
import tomllib
from collections.abc import Sequence
from typing import Any, Self

_logger = logging.getLogger(__name__)


@dataclasses.dataclass(frozen=True, kw_only=True)
class ProtobufConfig:
    """A configuration for the protobuf files.

    The configuration can be loaded from the `pyproject.toml` file using the class
    method `from_pyproject_toml()`.
    """

    proto_path: str = "proto"
    """The path of the root directory containing the protobuf files."""

    proto_glob: str = "*.proto"
    """The glob pattern to use to find the protobuf files."""

    include_paths: Sequence[str] = (
        "submodules/api-common-protos",
        "submodules/frequenz-api-common/proto",
    )
    """The paths to add to the include path when compiling the protobuf files."""

    py_path: str = "py"
    """The path of the root directory where the Python files will be generated."""

    docs_path: str = "protobuf-reference"
    """The path of the root directory where the documentation files will be generated."""

    @classmethod
    def from_pyproject_toml(
        cls, /, path: str = "pyproject.toml", **defaults: Any
    ) -> Self:
        """Create a new configuration by loading the options from a `pyproject.toml` file.

        The options are read from the `[tool.frequenz-repo-config.protobuf]`
        section of the `pyproject.toml` file.

        Args:
            path: The path to the `pyproject.toml` file.
            **defaults: The default values for the options missing in the file.  If
                a default is missing too, then the default in this class will be used.

        Returns:
            The configuration.
        """
        try:
            with pathlib.Path(path).open("rb") as toml_file:
                pyproject_toml = tomllib.load(toml_file)
        except FileNotFoundError:
            return cls(**defaults)
        except (IOError, OSError) as err:
            _logger.warning("WARNING: Failed to load pyproject.toml: %s", err)
            return cls(**defaults)

        try:
            config = pyproject_toml["tool"]["frequenz-repo-config"]["protobuf"]
        except KeyError:
            return cls(**defaults)

        known_keys = frozenset(defaults.keys())
        config_keys = frozenset(config.keys())
        if unknown_keys := config_keys - known_keys:
            _logger.warning(
                "WARNING: There are some configuration keys in pyproject.toml we don't "
                "know about and will be ignored: %s",
                ", ".join(f"'{k}'" for k in unknown_keys),
            )

        attrs = dict(defaults, **{k: config[k] for k in (known_keys & config_keys)})
        return cls(**attrs)



# License: MIT
# Copyright © 2023 Frequenz Energy-as-a-Service GmbH

"""Test utilities."""

MSG_UNEXPECTED_REPO_TYPES = """\
Missing (or unexpected) repository types. Please make sure update:

- the documentation (in particular in src/frequenz/repo/config/__init__.py)
- the code (including all packages)
- the tests
"""



# License: MIT
# Copyright © 2023 Frequenz Energy-as-a-Service GmbH

"""Generate the code reference pages."""

from frequenz.repo.config import mkdocs

mkdocs.generate_python_api_pages("src", "reference")



# License: MIT
# Copyright © 2023 Frequenz Energy-as-a-Service GmbH

"""Tests for the pyproject.toml file."""

import tomllib

from frequenz.repo.config import RepositoryType

from . import utils


def test_optional_dependencies() -> None:
    """Test that all repository types handled match the defined in RepositoryType."""
    expected = {t.value for t in RepositoryType}

    with open("pyproject.toml", "rb") as pyproject_file:
        pyproject_toml = tomllib.load(pyproject_file)
    defined = {
        k
        for k in pyproject_toml["project"]["optional-dependencies"].keys()
        if k != "dev" and not k.startswith("dev-")
    }
    assert defined == expected, utils.MSG_UNEXPECTED_REPO_TYPES



# License: MIT
# Copyright © 2023 Frequenz Energy-as-a-Service GmbH

"""Base types used accross the project."""

import enum as _enum


class RepositoryType(_enum.Enum):
    """Supported types of repository."""

    ACTOR = "actor"
    """SDK actor repository."""

    API = "api"
    """gRPC API repository."""

    APP = "app"
    """SDK application repository."""

    LIB = "lib"
    """General purpose library repository."""

    MODEL = "model"
    """SDK machine learning model repository."""



# License: MIT
# Copyright © 2023 Frequenz Energy-as-a-Service GmbH

"""General purpose utilities.

This module contains general purpose utilities that are used by the other
modules in this package.
"""


import pathlib as _pathlib
import tomllib as _tomllib
from collections.abc import Iterable, Mapping
from typing import TypeVar

_T = TypeVar("_T")


def flatten(iterables: Iterable[Iterable[_T]], /) -> Iterable[_T]:
    """Flatten an iterable of iterables into one iterable with all the elements.

    Args:
        iterables: The iterables to flatten.

    Returns:
        The flattened iterable.

    Example:
        >>> assert list(flatten([(1, 2), (3, 4)]) == [1, 2, 3, 4]
    """
    return (item for sublist in iterables for item in sublist)


def replace(iterable: Iterable[_T], replacements: Mapping[_T, _T], /) -> Iterable[_T]:
    """Replace elements in an iterable.

    Args:
        iterable: The iterable to replace elements in.
        replacements: A mapping of elements to replace with other elements.

    Yields:
        The next element in the iterable, with the replacements applied.

    Example:
        >>> assert list(replace([1, 2, 3], {1: 4, 2: 5})) == [4, 5, 3]
    """
    for item in iterable:
        if item in replacements:
            yield replacements[item]
        else:
            yield item


def deduplicate(iterable: Iterable[_T], /) -> Iterable[_T]:
    """Filter out duplicates from an iterable preserving the original iterable order.

    Args:
        iterable: The iterable to remove duplicates from.

    Returns:
        The elements of `iterable`, without duplicates but preserving order.
    """
    # We can't use a set() here because sets don't preserve order.  We use this hack
    # with dict.fromkeys() because dicts do preserve order in Python 3.7+.
    return dict.fromkeys(iterable).keys()


def existing_paths(paths: Iterable[str], /) -> Iterable[_pathlib.Path]:
    """Filter paths to only leave valid paths that exist and are unique.

    Args:
        paths: The paths to check and filter.

    Returns:
        An iterable with the valid paths as `pathlib.Path` objects.

    Example:
        >>> assert list(existing_paths([".", "/fake"])) == [pathlib.Path(".")]
    """
    return deduplicate(p for p in map(_pathlib.Path, paths) if p.exists())


def is_python_file(path: _pathlib.Path, /) -> bool:
    """Check if a path is a Python file.

    Args:
        path: The path to check.

    Returns:
        `True` if the path is a Python file, `False` otherwise.
    """
    return path.suffix in (".py", ".pyi")


def path_to_package(path: _pathlib.Path, root: _pathlib.Path | None = None) -> str:
    """Convert paths to Python package names.

    Paths should exist and be either a directory or a file ending with `.pyi?`
    (otherwise this function will assert). The `root` and `path` are
    concatenated when performing the check.

    Directory separators in `path` are replaced with `.` and the `.pyi?` suffix
    is removed (if present).

    Args:
        path: The path to convert.
        root: The root where the path is located. If `None`, then it is
            considered present in the current working directory.

    Returns:
        The package name based on `path`.

    Examples:
        * `src/frequenz/pkg` (`root="src"`) will be converted to `frequenz.pkg`.
        * `noxfile.py` (without `root`) will be converted to `noxfile`.
    """
    real_path = path
    if root is not None:
        real_path = root / path
    assert real_path.is_dir() or is_python_file(real_path)

    if is_python_file(real_path):
        path = path.with_suffix("")
    return path.as_posix().replace("/", ".")


def find_toplevel_package_dirs(
    path: _pathlib.Path, /, *, root: _pathlib.Path | None = None
) -> Iterable[_pathlib.Path]:
    """Find top-level packages directories in a `path`.

    Searches recursively for the top-level packages in `path`, relative to
    `root`.

    Args:
        path: The path to look for python packages.
        root: The part of the path that is considered the root and will be
            removed from the resulting path. If `None` then `path` is used as
            `root`.

    Returns:
        The top-level paths that contains a `__init__.py` file, with `root`
        removed.

    Examples:
        If we have a directory like the following:

        ```
        .
        ├── noxfile.py
        └── src
            └── frequenz
                └── repo
                    └── config
                        ├── __init__.py
                        ├── nox
                        │   ├── config.py
                        │   ├── default.py
                        │   ├── __init__.py
                        │   ├── session.py
                        │   └── util.py
                        └── setuptools.py
        ```

        Then calling `find_toplevel_package_dirs(pathlib.Path("src"))` will
        return an iterator producing: `["frequenz/repo/config"]`.
    """
    if root is None:
        root = path
    # Bail out early if it is a directory with a __init__.py to avoid getting
    # sub-packages
    if (path / "__init__.py").exists():
        return [path.relative_to(root)]
    if path.is_dir():
        return flatten(
            [find_toplevel_package_dirs(p, root=root) for p in path.iterdir()]
        )
    return ()


def min_dependencies() -> list[str]:
    """Extract the minimum dependencies from pyproject.toml.

    Returns:
        The minimun dependencies defined in pyproject.toml.

    Raises:
        RuntimeError: If minimun dependencies are not properly set in pyproject.toml.
    """
    with open("pyproject.toml", "rb") as toml_file:
        data = _tomllib.load(toml_file)

    min_deps: list[str] = []

    dependencies = data.get("project", {}).get("dependencies", {})
    if not dependencies:
        return min_deps

    for dep in dependencies:
        min_dep = dep.split(",")[0]
        if any(op in min_dep for op in (">=", "==")):
            min_deps.append(min_dep.replace(">=", "=="))
        else:
            raise RuntimeError(f"Minimum requirement is not set: {dep}")
    return min_deps


def discover_paths() -> list[str]:
    """Discover paths to check.

    Discover the paths to check by looking into different sources, like the
    `pyproject.toml` file.

    Currently the following paths are discovered:

    - The `testpaths` option in the `tools.pytest.ini_options` section of
      `pyproject.toml`.

    Returns:
        The discovered paths to check.
    """
    with open("pyproject.toml", "rb") as toml_file:
        data = _tomllib.load(toml_file)

    testpaths: list[str] = (
        data.get("tool", {})
        .get("pytest", {})
        .get("ini_options", {})
        .get("testpaths", [])
    )

    return list(deduplicate(testpaths))



# License: MIT
# Copyright © 2023 Frequenz Energy-as-a-Service GmbH

r"""Frequenz project setup tools and common configuration.

The tools are provided to configure the main types of repositories most commonly used at
Frequenz, defined in
[`frequenz.repo.config.RepositoryType`][].

- actor: SDK actors
- api: gRPC APIs
- app: SDK applications
- lib: General purpose Python libraries
- model: SDK machine learning models

# Common

## `nox` (running tests and linters)

### Writing the `noxfile.py`

Projects wanting to use `nox` to run lint checkers and other utilities can use
the [`frequenz.repo.config.nox`][] package.

When writing the `noxfile.py` you should import the `nox` module from this
package and use the [`frequenz.repo.config.nox.configure`][] function,
which will configure all nox sessions.

To use the default options, you should call `configure()` using one of the [repository
types][frequenz.repo.config.RepositoryType].  For example:

```python
from frequenz.repo.config import RepositoryType, nox

nox.configure(RepositoryType.LIB)
```

Again, make sure to pick the correct project typedefault configuration based on the type of your
project (`actor_config`, `api_config`, `app_config`, `lib_config`, `model_config`).

If you need to use some custom configuration, you can start from the default settings in
the [`frequenz.repo.config.nox.default`][] module,
[copying][frequenz.repo.config.nox.config.Config.copy] it and changing whatever you
need to customize.  For example:

```python
from frequenz.repo.config import nox
from frequenz.repo.config.nox import default

config = default.lib_config.copy()
config.opts.black.append("--diff")
nox.configure(config)
```

If you need further customization or to define new sessions, you can use the
following modules:

- [`frequenz.repo.config.nox.config`][]: Low-level utilities to configure nox
  sessions. It defines the `Config` and CommandsOptions` classes and the actual
  implementation of the `configure()` function. It also defines the `get()`
  function, which can be used to get the currently used configuration object.

- [`frequenz.repo.config.nox.session`][]: Predefined nox sessions. These are
  the sessions that are used by default.

- [`frequenz.repo.config.nox.util`][]: General purpose utility functions.

### `pyproject.toml` configuration

All sessions configured by this package expect the `pyproject.toml` file to
define specific *dev* dependencies that will be used by the different `nox`
sessions.

The following optional dependencies are used and must be defined:

- `dev-docstrings`: Dependencies to lint the documentation.

  At least these packages should be included:

  - `pydocstyle`: To check the docstrings' format.
  - `darglint`: To check the docstrings' content.

- `dev-formatting`: Dependencies to check the code's formatting.

  At least these packages should be included:

  - `black`: To check the code's formatting.
  - `isort`: To check the imports' formatting.

- `dev-mypy`: Dependencies to run `mypy` to check the code's type annotations.

  At least these packages should be included:

  - `mypy`: To check the code's type annotations.

- `dev-pylint`: Dependencies to run `pylint` to lint the code.

  At least these packages should be included:

  - `pylint`: To lint the code.

- `dev-pytest`: Dependencies to run the tests using `pytest`.

  At least these packages should be included:

  - `pytest`: To run the tests.

For some of these you should install too any other dependencies that are used
by the project. For example, if the project uses `pytest-asyncio`, you should
include it in the `dev-pytest` optional dependency.

It is also recommended, but not required, to provide a global `dev` optional
dependency that includes all the other optional dependencies, so users can
install all the dependencies needed while developing the project without having
to run `nox`, which might be a bit slow if you want to do quick iterations.

```console
$ pip install -e .[dev]
...
$ pytest
...
```

Here is a sample `pyproject.toml` file that defines all the optional
dependencies:

```toml
[project]
name = "my-package"
# ...

[project.optional-dependencies]
dev-docstrings = ["pydocstyle == 6.3.0", "darglint == 1.8.1"]
dev-formatting = ["black == 23.3.0", "isort == 5.12.0"]
dev-mkdocs = [
  "mike == 1.1.2",
  "mkdocs-gen-files == 0.5.0",
  "mkdocs-literate-nav == 0.6.0",
  "mkdocs-material == 9.1.16",
  "mkdocs-section-index == 0.3.5",
  "mkdocstrings[python] == 0.22.0",
]
dev-mypy = [
  "mypy == 1.1.1",
  # For checking tests
  "my-package[dev-mkdocs,dev-pytest]",
]
dev-pylint = [
  "pylint == 2.17.1",
  "pylint-google-style-guide-imports-enforcing == 1.3.0",
  # For checking tests
  "my-package[dev-mkdocs,dev-pytest]",
]
dev-pytest = [
  "pytest == 7.2.2",
  "pytest-asyncio == 0.21.0",
  "pytest-mock == 3.10.0",
]
dev = [
  "my-package[dev-mkdocs,dev-docstrings,dev-formatting,dev-mypy,dev-nox,dev-pylint,dev-pytest]",
]
```

## `mkdocs` (generating documentation)

### API reference generation

The API documnentation can be automatically generated from the source files using the
[`frequenz.repo.config.mkdocs`][] package as when run as a
[`mkdocs-gen-files`](https://oprypin.github.io/mkdocs-gen-files/) plugin script.

To enable it you just need to make sure the `mkdocs-gen-files`, `mkdocs-literate-nav`
and `mkdocstrings[python]` packages are installed (look at the `pyproject.toml`
configuration in the `nox` section) and add the following configuration to the
`mkdocs.yml` file:

```yaml
plugins:
  - gen-files:
      scripts:
        - path/to/my/custom/script.py
```

By default this script will look for files in the `src/` directory and generate the
documentation files in the `python-reference/` directory inside `mkdocs` output directory
(`site` by defaul).

If you need to customize the above paths, you can create a new script to use with the
`mkdocs-gen-files` plugin as follows:

```python
from frequenz.repo.config import mkdocs

mkdocs.generate_python_api_pages("my_sources", "API")
```

Where `my_sources` is the directory containing the source files and `API` is the
directory where to generate the documentation files (relative to `mkdocs` output
directory).

And then replace this configuration in the `mkdocs.yml` file:

```yaml
plugins:
  - gen-files:
      scripts:
        - path/to/my/custom/script.py
```

# APIs

## Protobuf configuation

Support is provided to generate files from *protobuf* files.  To do this, it is possible
to configure the options to use while generating the files for different purposes
(language bindings, documentation, etc.).

The configuration can be done in the `pyproject.toml` file as follows:

```toml
[tool.frequenz_repo_config.protobuf]
# Location of the proto files relative to the root of the repository (default: "proto")
proto_path = "proto_files"
# Glob pattern to use to find the proto files in the proto_path (default: "*.proto")
proto_glob = "*.prt"  # Default: "*.proto"
# List of paths to pass to the protoc compiler as include paths (default:
# ["submodules/api-common-protos", "submodules/frequenz-api-common/proto"])
include_paths = ["submodules/api-common-protos"]
# Path where to generate the Python files (default: "py")
py_path = "generated"
# Path where to generate the documentation files (default: "protobuf-reference")
docs_path = "API"
```

If the defaults are not suitable for you (for example you need to use more or less
submodules or your proto files are located somewhere else), please adjust the
configuration to match your project structure.

### `mkdocs` API reference generation

If your project provides *protobuf* files, you can also generate the API
documentation for them adding one more line to the script provided in the common
section:

```python
from frequenz.repo.config import mkdocs

mkdocs.generate_python_api_pages("my_sources", "API-py")
mkdocs.generate_protobuf_api_pages()
```

This will use the configuration in the `pyproject.toml` file and requires `docker` to
run (it uses the `pseudomuto/protoc-gen-doc` docker image.

### `setuptools` gRPC support

When configuring APIs it is assumed that they have a gRPC interface.

The project structure is assumed to be as described in the *Protobuf configuration*
section plus the following:

- `pytests/`: Directory containing the tests for the Python code.
- `submodules/api-common-protos`: Directory containing the Git submodule with the
  `google/api-common-protos` repository.
- `submodules/frequenz-api-common`: Directory containing the Git submodule with the
  `frequenz-floss/frequenz-api-common` repository.

Normally Frequenz APIs use basic types from
[`google/api-common-protos`](https://github.com/googleapis/api-common-protos) and
[`frequenz-floss/frequenz-api-common`](https://github.com/frequenz-floss/frequenz-api-common),
so you need to make sure the proper submodules are added to your project:

```sh
mkdir submodules
git submodule add https://github.com/googleapis/api-common-protos.git \
        submodules/api-common-protos
git submodule add https://github.com/frequenz-floss/frequenz-api-common.git \
        submodules/frequenz-api-common
git commit -m "Add api-common-protos and frequenz-api-common submodules" submodules
```

Then you need to add this package as a build dependency and a few extra
dependencies to your project, for example:

```toml
requires = [
  "setuptools >= 67.3.2, < 68",
  "setuptools_scm[toml] >= 7.1.0, < 8",
  "frequenz-repo-config[api] >= 0.1.0, < 0.2.0",
]
build-backend = "setuptools.build_meta"

[project]
dependencies = [
  "frequenz-api-common >= 0.2.0, < 0.3.0",
  "googleapis-common-protos >= 1.56.2, < 2",
  "grpcio >= 1.51.1, < 2",
]
```

Note the `api` extra in `frequenz-repo-config[api]`, this will ensure all
dependencies to build the protocol files will be installed when building the
package. Of course you need to replace the version numbers with the correct
ones too.

You should also add the following configuration to your `pyproject.toml` file
to make sure the generated files are included in the wheel:

```toml
[tool.setuptools.package-dir]
"" = "py"

[tool.setuptools.package-data]
"*" = ["*.pyi"]

[tools.pytest.ini_options]
testpaths = ["pytests"]
```

Finally you need to make sure to include the generated `*.pyi` files in the
source distribution, as well as the Google api-common-protos files, as it
is not handled automatically yet
([#13](https://github.com/frequenz-floss/frequenz-repo-config-python/issues/13)).
Make sure to include these lines in the `MANIFEST.in` file:

```
recursive-include submodules/api-common-protos/google *.proto
recursive-include submodules/frequenz-api-common/proto *.proto
```

Please adapt the instructions above to your project structure if you need to change the
defaults.
"""

from ._core import RepositoryType

__all__ = [
    "RepositoryType",
]



# License: MIT
# Copyright © {% now 'utc', '%Y' %} {{cookiecutter.author_name}}

"""Configuration file for nox."""

from frequenz.repo.config import RepositoryType, nox

nox.configure(RepositoryType.{{cookiecutter.type | upper}})



# License: MIT
# Copyright © 2023 Frequenz Energy-as-a-Service GmbH

"""Predefined nox sessions.

This module defines the predefined nox sessions that are used by the default.
"""

import nox

from . import config as _config
from . import util as _util


@nox.session
def ci_checks_max(session: nox.Session) -> None:
    """Run all checks with max dependencies in a single session.

    This is faster than running the checks separately, so it is suitable for CI.

    This does NOT run pytest_min, so that needs to be run separately as well.

    Args:
        session: the nox session.
    """
    session.install("-e", ".[dev]")

    formatting(session, False)
    mypy(session, False)
    pylint(session, False)
    docstrings(session, False)
    pytest_max(session, False)


@nox.session
def formatting(session: nox.Session, install_deps: bool = True) -> None:
    """Check code formatting with black and isort.

    Args:
        session: the nox session.
        install_deps: True if dependencies should be installed.
    """
    if install_deps:
        session.install("-e", ".[dev-formatting]")

    conf = _config.get()
    session.run("black", *conf.opts.black, *conf.path_args(session))
    session.run("isort", *conf.opts.isort, *conf.path_args(session))


@nox.session
def mypy(session: nox.Session, install_deps: bool = True) -> None:
    """Check type hints with mypy.

    Args:
        session: the nox session.
        install_deps: True if dependencies should be installed.
    """
    if install_deps:
        # install the package itself as editable, so that it is possible to do
        # fast local tests with `nox -R -e mypy`.
        session.install("-e", ".[dev-mypy]")

    conf = _config.get()
    pkg_args = _util.flatten(("-p", p) for p in conf.package_args(session))
    session.run("mypy", *conf.opts.mypy, *pkg_args)


@nox.session
def pylint(session: nox.Session, install_deps: bool = True) -> None:
    """Check for code smells with pylint.

    Args:
        session: the nox session.
        install_deps: True if dependencies should be installed.
    """
    if install_deps:
        # install the package itself as editable, so that it is possible to do
        # fast local tests with `nox -R -e pylint`.
        session.install("-e", ".[dev-pylint]")

    conf = _config.get()
    session.run("pylint", *conf.opts.pylint, *conf.path_args(session))


@nox.session
def docstrings(session: nox.Session, install_deps: bool = True) -> None:
    """Check docstring tone with pydocstyle and param descriptions with darglint.

    Args:
        session: the nox session.
        install_deps: True if dependencies should be installed.
    """
    if install_deps:
        session.install("-e", ".[dev-docstrings]")

    conf = _config.get()
    session.run("pydocstyle", *conf.opts.pydocstyle, *conf.path_args(session))

    # Darglint checks that function argument and return values are documented.
    # This is needed only for the `src` dir, so we exclude the other top level
    # dirs that contain code, unless some paths were specified by argument, in
    # which case we use those untouched.
    darglint_paths = session.posargs or filter(
        # pylint: disable=fixme
        # TODO: Make these exclusions configurable
        lambda path: not (path.startswith("tests") or path.startswith("benchmarks")),
        conf.path_args(session),
    )
    session.run("darglint", *conf.opts.darglint, *darglint_paths)


@nox.session
def pytest_max(session: nox.Session, install_deps: bool = True) -> None:
    """Test the code against max dependency versions with pytest.

    Args:
        session: the nox session.
        install_deps: True if dependencies should be installed.
    """
    if install_deps:
        # install the package itself as editable, so that it is possible to do
        # fast local tests with `nox -R -e pytest_max`.
        session.install("-e", ".[dev-pytest]")

    _pytest_impl(session, "max")


@nox.session
def pytest_min(session: nox.Session, install_deps: bool = True) -> None:
    """Test the code against min dependency versions with pytest.

    Args:
        session: the nox session.
        install_deps: True if dependencies should be installed.
    """
    if install_deps:
        # install the package itself as editable, so that it is possible to do
        # fast local tests with `nox -R -e pytest_min`.
        session.install("-e", ".[dev-pytest]", *_util.min_dependencies())

    _pytest_impl(session, "min")


def _pytest_impl(
    session: nox.Session, max_or_min_deps: str  # pylint: disable=unused-argument
) -> None:
    conf = _config.get()
    session.run("pytest", *conf.opts.pytest, *session.posargs)

    # pylint: disable=fixme
    # TODO: Implement coverage reporting, we need to research this a bit and it
    # makes sense to do so when we actually collect the coverage somewhere
    # "--cov=frequenz.sdk",
    # "--cov-report=term",
    # f"--cov-report=html:.htmlcov-{max_or_min_deps}",



# License: MIT
# Copyright © {% now 'utc', '%Y' %} {{cookiecutter.author_name}}

"""{{cookiecutter.description}}.

TODO(cookiecutter): Add a more descriptive module description.
"""
{%- if cookiecutter.type != "api" %}


# TODO(cookiecutter): Remove this function
def delete_me(*, blow_up: bool = False) -> bool:
    """Do stuff for demonstration purposes.

    Args:
        blow_up: If True, raise an exception.

    Returns:
        True if no exception was raised.

    Raises:
        RuntimeError: if blow_up is True.
    """
    if blow_up:
        raise RuntimeError("This function should be removed!")
    return True
{%- endif %}



# License: MIT
# Copyright © 2023 Frequenz Energy-as-a-Service GmbH

"""Configuration utilities for nox.

This module provides utilities to configure the nox sessions. It provides a
`Config` and a `CommandsOptions` class, which are used to configure the nox sessions.

The `get()` function can be used to retrieve the current configuration object
so it can be used when implementing custom nox sessions.

The `configure()` function must be called before `get()` is used.
"""

import dataclasses as _dataclasses
import itertools as _itertools
from typing import Self, assert_never, overload

import nox as _nox

from .._core import RepositoryType
from . import util as _util


@_dataclasses.dataclass(kw_only=True, slots=True)
class CommandsOptions:
    """Command-line options for each command."""

    black: list[str] = _dataclasses.field(default_factory=lambda: [])
    """Command-line options for the `black` command."""

    darglint: list[str] = _dataclasses.field(default_factory=lambda: [])
    """Command-line options for the `darglint` command."""

    isort: list[str] = _dataclasses.field(default_factory=lambda: [])
    """Command-line options for the `isort` command."""

    mypy: list[str] = _dataclasses.field(default_factory=lambda: [])
    """Command-line options for the `mypy` command."""

    pydocstyle: list[str] = _dataclasses.field(default_factory=lambda: [])
    """Command-line options for the `pydocstyle` command."""

    pylint: list[str] = _dataclasses.field(default_factory=lambda: [])
    """Command-line options for the `pylint` command."""

    pytest: list[str] = _dataclasses.field(default_factory=lambda: [])
    """Command-line options for the `pytest` command."""

    def copy(self) -> Self:
        """Create a new object as a copy of self.

        Returns:
            The copy of self.
        """
        return _dataclasses.replace(self)


@_dataclasses.dataclass(kw_only=True, slots=True)
class Config:
    """Configuration for nox sessions."""

    opts: CommandsOptions = _dataclasses.field(default_factory=CommandsOptions)
    """Command-line options for each command used by sessions."""

    sessions: list[str] = _dataclasses.field(default_factory=lambda: [])
    """List of sessions to run."""

    source_paths: list[str] = _dataclasses.field(default_factory=lambda: [])
    """List of paths containing source files that should be analyzed by the sessions.

    Source paths are inspected for `__init__.py` files to look for packages.
    The path should be the top-level directory containing packages that will be
    actually part of the distribution, not development paths, like tests,
    benchmarks, etc.

    This path will be removed when calculating the package name for the found
    packages. `mypy` needs the package name to be able to do full import
    checking.
    """

    extra_paths: list[str] = _dataclasses.field(default_factory=lambda: [])
    """List of extra paths to be analyzed by the sessions.

    These are not inspected for packages, as they are passed verbatim to the
    tools invoked by the sessions.
    """

    def __post_init__(self) -> None:
        """Initialize the configuration object.

        This will add extra paths discovered in config files and other sources.
        """
        for path in _util.discover_paths():
            if path not in self.extra_paths:
                self.extra_paths.append(path)

    def copy(self, /) -> Self:
        """Create a new object as a copy of self.

        Returns:
            The copy of self.
        """
        return _dataclasses.replace(self)

    def path_args(self, session: _nox.Session, /) -> list[str]:
        """Return the file paths to run the checks on.

        If positional arguments are present in the nox session, those are used
        as the file paths verbatim, and if not, all **existing** `source_paths`
        and `extra_paths` are used.

        Args:
            session: The nox session to use to look for command-line arguments.

        Returns:
            The file paths to run the checks on.
        """
        if session.posargs:
            return session.posargs

        return list(
            str(p) for p in _util.existing_paths(self.source_paths + self.extra_paths)
        )

    def package_args(self, session: _nox.Session, /) -> list[str]:
        """Return the package names to run the checks on.

        If positional arguments are present in the nox session, those are used
        as the file paths verbatim, and if not, all **existing** `source_paths`
        are searched for python packges by looking for `__init__.py` files.
        `extra_paths` are used as is, only converting the paths to python
        package names (replacing `/` with `.` and removing the suffix `.pyi?`
        if it exists.

        Args:
            session: The nox session to use to look for command-line arguments.

        Returns:
            The package names found in the `source_paths`.
        """
        if session.posargs:
            return session.posargs

        sources_package_dirs_with_roots = (
            (p, _util.find_toplevel_package_dirs(p))
            for p in _util.existing_paths(self.source_paths)
        )

        source_packages = (
            _util.path_to_package(pkg_path, root=root)
            for root, pkg_paths in sources_package_dirs_with_roots
            for pkg_path in pkg_paths
        )

        extra_packages = (
            _util.path_to_package(p) for p in _util.existing_paths(self.extra_paths)
        )

        return list(
            _util.deduplicate(_itertools.chain(source_packages, extra_packages))
        )


_config: Config | None = None
"""The global configuration object."""


def get() -> Config:
    """Get the global configuration object.

    This will assert if `configure()` wasn't called before.

    Returns:
        The global configuration object.
    """
    assert _config is not None, "You must call configure() before using this function"
    return _config


@overload
def configure(conf: Config, /, *, import_default_sessions: bool = True) -> None:
    """Configure nox using the provided configuration.

    Args:
        conf: The configuration to use to configure nox.
        import_default_sessions: Whether to import the default sessions or not.
            This is only necessary if you want to avoid using the default provided
            sessions and use your own.
    """


@overload
def configure(
    repo_type: RepositoryType, /, *, import_default_sessions: bool = True
) -> None:
    """Configure nox using the provided repository type.

    Args:
        repo_type: The repository type to use to configure nox.  This will use the
            default configuration in [`frequenz.repo.config.nox.default`][] for that
            type of repository.
        import_default_sessions: Whether to import the default sessions or not.
            This is only necessary if you want to avoid using the default provided
            sessions and use your own.
    """


def configure(
    conf: Config | RepositoryType, /, *, import_default_sessions: bool = True
) -> None:
    """Configure nox using the provided configuration or repository type.

    Args:
        conf: The configuration to use to configure nox, or the repository type to use
            to configure nox.  The later will use the default configuration in
            [`frequenz.repo.config.nox.default`][] for that type of repository.
        import_default_sessions: Whether to import the default sessions or not.
            This is only necessary if you want to avoid using the default provided
            sessions and use your own.
    """
    global _config  # pylint: disable=global-statement

    # We need to make sure sessions are imported, otherwise they won't be visible to nox.
    if import_default_sessions:
        # pylint: disable=import-outside-toplevel,cyclic-import
        from . import session as _

    match conf:
        case Config():
            _config = conf
        case RepositoryType() as repo_type:
            # pylint: disable=import-outside-toplevel,cyclic-import
            from . import default

            match repo_type:
                case RepositoryType.ACTOR:
                    _config = default.actor_config
                case RepositoryType.API:
                    _config = default.api_config
                case RepositoryType.APP:
                    _config = default.app_config
                case RepositoryType.LIB:
                    _config = default.lib_config
                case RepositoryType.MODEL:
                    _config = default.model_config
                case _ as unhandled:
                    assert_never(unhandled)

    _nox.options.sessions = _config.sessions



# License: MIT
# Copyright © 2023 Frequenz Energy-as-a-Service GmbH

"""Utilities to build noxfiles.

The main entry point is the [`configure()`][frequenz.repo.config.nox.configure]
function, which will configure all nox sessions according to some configuration.

To use the default options, you should call `configure()` using one of the [repository
types][frequenz.repo.config.RepositoryType].  For example:

```python
from frequenz.repo.config import RepositoryType, nox

nox.configure(RepositoryType.LIB)
```

Again, make sure to pick the correct project typedefault configuration based on the type
of your project (`actor_config`, `api_config`, `app_config`, `lib_config`,
`model_config`).

If you need to use some custom configuration, you can start from the default settings in
the [`frequenz.repo.config.nox.default`][] module,
[copying][frequenz.repo.config.nox.config.Config.copy] it and changing whatever you
need to customize.  For example:

```python
from frequenz.repo.config import nox
from frequenz.repo.config.nox import default

config = default.lib_config.copy()
config.opts.black.append("--diff")
nox.configure(config)
```

If you need further customization or to define new sessions, you can use the following
modules:

- [`frequenz.repo.config.nox.config`][]: Low-level utilities to configure nox sessions.
  It defines the `Config` and CommandsOptions` classes and the actual implementation of
  the `configure()` function. It also defines the `get()` function, which can be used to
  get the currently used configuration object.

- [`frequenz.repo.config.nox.session`][]: Predefined nox sessions. These are the
  sessions that are used by default.

- [`frequenz.repo.config.nox.util`][]: General purpose utility functions.
"""

from .config import configure

__all__ = [
    "configure",
]



# License: MIT
# Copyright © 2023 Frequenz Energy-as-a-Service GmbH

"""Setuptool hooks to build protobuf files.

This module contains a setuptools command that can be used to compile protocol
buffer files in a project.

It also runs the command as the first sub-command for the build command, so
protocol buffer files are compiled automatically before the project is built.
"""

import pathlib as _pathlib
import subprocess as _subprocess
import sys as _sys
import tomllib as _tomllib

import setuptools as _setuptools
import setuptools.command.build as _build_command

from .. import protobuf as _protobuf


class CompileProto(_setuptools.Command):
    """Build the Python protobuf files."""

    proto_path: str
    """The path of the root directory containing the protobuf files."""

    proto_glob: str
    """The glob pattern to use to find the protobuf files."""

    include_paths: str
    """Comma-separated list of paths to include when compiling the protobuf files."""

    py_path: str
    """The path of the root directory where the Python files will be generated."""

    description: str = "compile protobuf files"
    """Description of the command."""

    user_options: list[tuple[str, str | None, str]] = [
        (
            "proto-path=",
            None,
            "path of the root directory containing the protobuf files",
        ),
        ("proto-glob=", None, "glob pattern to use to find the protobuf files"),
        (
            "include-paths=",
            None,
            "comma-separated list of paths to include when compiling the protobuf files",
        ),
        (
            "py-path=",
            None,
            "path of the root directory where the Python files will be generated",
        ),
    ]
    """Options of the command."""

    def initialize_options(self) -> None:
        """Initialize options."""
        config = _protobuf.ProtobufConfig.from_pyproject_toml()

        self.proto_path = config.proto_path
        self.proto_glob = config.proto_glob
        self.include_paths = ",".join(config.include_paths)
        self.py_path = config.py_path

    def finalize_options(self) -> None:
        """Finalize options."""

    def run(self) -> None:
        """Compile the Python protobuf files."""
        include_paths = self.include_paths.split(",")
        proto_files = [
            str(p) for p in _pathlib.Path(self.proto_path).rglob(self.proto_glob)
        ]

        if not proto_files:
            print(
                f"No proto files found in {self.proto_path}/**/{self.proto_glob}/, "
                "skipping compilation of proto files."
            )
            return

        protoc_cmd = (
            [_sys.executable, "-m", "grpc_tools.protoc"]
            + [f"-I{p}" for p in [*include_paths, self.proto_path]]
            + [
                f"--{opt}={self.py_path}"
                for opt in "python_out grpc_python_out mypy_out mypy_grpc_out".split()
            ]
            + proto_files
        )

        print(f"Compiling proto files via: {' '.join(protoc_cmd)}")
        _subprocess.run(protoc_cmd, check=True)


# This adds the compile_proto command to the build sub-command.
# The name of the command is mapped to the class name in the pyproject.toml file,
# in the [project.entry-points.distutils.commands] section.
# The None value is an optional function that can be used to determine if the
# sub-command should be executed or not.
_build_command.build.sub_commands.insert(0, ("compile_proto", None))



# License: MIT
# Copyright © 2023 Frequenz Energy-as-a-Service GmbH

"""Default nox configuration for different types of repositories.

This module provides the default configuration for the different types of
repositories defined by
[`frequenz.repo.config.RepositoryType`][frequenz.repo.config.RepositoryType].

The `actor_config`, `api_config`, `app_config`, `lib_config`, and `model_config`
variables are the default configurations for libraries, APIs, actors and applications,
respectively. The `common_config` variable is the default configuration for all types of
repositories.

The `actor_command_options`, `api_command_options`, `app_command_options`,
`lib_command_options`, and `model_command_options` variables are the default
command-line options for the same types of repositories, and the
`common_command_options` variable is the default command-line options for all types of
repositories.

They can be modified before being passed to
[`nox.configure()`][frequenz.repo.config.nox.configure] by using the
[`CommandsOptions.copy()`][frequenz.repo.config.nox.config.CommandsOptions.copy]
method.
"""

import dataclasses

from . import config as _config
from . import util as _util

common_command_options: _config.CommandsOptions = _config.CommandsOptions(
    black=[
        "--check",
    ],
    darglint=[
        "-v2",  # for verbose error messages.
    ],
    isort=[
        "--diff",
        "--check",
    ],
    mypy=[
        "--install-types",
        "--namespace-packages",
        "--non-interactive",
        "--explicit-package-bases",
        "--strict",
    ],
    # SDK: pylint "--extension-pkg-whitelist=pydantic"
    pytest=[
        "-W=all",
        "-vv",
    ],
)
"""Default command-line options for all types of repositories."""

common_config = _config.Config(
    opts=common_command_options.copy(),
    sessions=[
        "formatting",
        "mypy",
        "pylint",
        "docstrings",
        "pytest_min",
        "pytest_max",
    ],
    source_paths=[
        "src",
    ],
    extra_paths=[
        "benchmarks",
        "docs",
        "examples",
        "noxfile.py",
        "tests",
    ],
)
"""Default configuration for all types of repositories."""

actor_command_options: _config.CommandsOptions = common_command_options.copy()
"""Default command-line options for actors."""

actor_config: _config.Config = common_config.copy()
"""Default configuration for actors."""

api_command_options: _config.CommandsOptions = common_command_options.copy()
"""Default command-line options for APIs."""

api_config: _config.Config = dataclasses.replace(
    common_config,
    opts=api_command_options,
    # We don't check the sources at all because they are automatically generated.
    source_paths=[],
    # We adapt the path to the tests.
    extra_paths=list(_util.replace(common_config.extra_paths, {"tests": "pytests"})),
)
"""Default configuration for APIs.

Same as `common_config`, but with `source_paths` replacing `"src"` with `"py"`
and `extra_paths` replacing `"tests"` with `"pytests"`.
"""

app_command_options: _config.CommandsOptions = common_command_options.copy()
"""Default command-line options for applications."""

app_config: _config.Config = common_config.copy()
"""Default configuration for applications."""

lib_command_options: _config.CommandsOptions = common_command_options.copy()
"""Default command-line options for libraries."""

lib_config: _config.Config = common_config.copy()
"""Default configuration for libraries."""

model_command_options: _config.CommandsOptions = common_command_options.copy()
"""Default command-line options for models."""

model_config: _config.Config = common_config.copy()
"""Default configuration for models."""



# License: MIT
# Copyright © 2023 Frequenz Energy-as-a-Service GmbH

"""Local cookiecutter extensions.

This module contains local cookiecutter extensions that are useful for generating the
project structure.
"""

import json as _json

from cookiecutter.utils import simple_filter as _simple_filter


def _build_identifier(repo_type: str, name: str, separator: str) -> str:
    """Build an identifier depending on the repository type.

    Args:
        repo_type: The repository type.
        name: The project name.
        separator: The separator to use to construct the identifier.

    Returns:
        The built identifier.
    """
    middle = f"{repo_type}{separator}" if repo_type != "lib" else ""
    return f"frequenz{separator}{middle}{name}"


def _get_from_json(key: str) -> str:
    """Get a string from the cookiecutter.json file.

    Args:
        key: The key to get the string for.

    Returns:
        The string from the cookiecutter.json file.
    """
    with open("../cookiecutter.json", encoding="utf8") as cookiecutter_json_file:
        return str(_json.load(cookiecutter_json_file)[key])


# Ignoring because cookiecutter simple_filter decorator is not typed.
@_simple_filter  # type: ignore[misc]
def python_package(cookiecutter: dict[str, str]) -> str:
    """Generate the Python package (import) depending on the repository type.

    Args:
        cookiecutter: The cookiecutter context.

    Returns:
        The Python package name that can be used in import statements.
    """
    return _build_identifier(cookiecutter["type"], cookiecutter["name"], ".")


@_simple_filter  # type: ignore[misc]
def pypi_package_name(cookiecutter: dict[str, str]) -> str:
    """Generate the PyPI package name depending on the repository type.

    Args:
        cookiecutter: The cookiecutter context.

    Returns:
        The PyPI package name depending on the repository type.
    """
    return _build_identifier(cookiecutter["type"], cookiecutter["name"], "-")


@_simple_filter  # type: ignore[misc]
def github_repo_name(cookiecutter: dict[str, str]) -> str:
    """Generate the Python package name depending on the repository type.

    Args:
        cookiecutter: The cookiecutter context.

    Returns:
        The Python package name depending on the repository type.
    """
    pypi = _build_identifier(cookiecutter["type"], cookiecutter["name"], "-")
    end = "-python" if cookiecutter["type"] == "lib" else ""
    return f"{pypi}{end}"


@_simple_filter  # type: ignore[misc]
def title(cookiecutter: dict[str, str]) -> str:
    """Build a default mkdocs site name for the project.

    Args:
        cookiecutter: The cookiecutter context.

    Returns:
        The default site name.
    """
    match cookiecutter["type"]:
        case "actor":
            return f"Frequenz {cookiecutter['name']} actor"
        case "api":
            return f"Frequenz {cookiecutter['name']} API"
        case "lib":
            return f"Freqenz {cookiecutter['name']} library"
        case "app":
            return f"Frequenz {cookiecutter['name']} app"
        case "model":
            return f"Frequenz {cookiecutter['name']} AI model"
        case _ as repo_type:
            assert False, f"Unhandled repository type {repo_type!r}"


@_simple_filter  # type: ignore[misc]
def src_path(cookiecutter: dict[str, str]) -> str:
    """Build a default source path for the project.

    Args:
        cookiecutter: The cookiecutter context.

    Returns:
        The default source path.
    """
    match cookiecutter["type"]:
        case "api":
            return "py"
        case "actor" | "lib" | "app" | "model":
            return "src"
        case _ as repo_type:
            assert False, f"Unhandled repository type {repo_type!r}"


@_simple_filter  # type: ignore[misc]
def keywords(cookiecutter: dict[str, str]) -> str:
    """Extend cookiecutter["keywords"] with predefined ones by repository type.

    Args:
        cookiecutter: The cookiecutter context.

    Returns:
        The extended keywords.
    """
    repo_type = cookiecutter["type"]
    extended_keywords = ["frequenz", "python", repo_type]
    if repo_type == "api":
        extended_keywords.extend(["grpc", "protobuf", "rpc"])
    extended_keywords.append(cookiecutter["name"])
    default = _get_from_json("keywords")
    cookiecutter_keywords = cookiecutter["keywords"]
    if cookiecutter_keywords == default:
        cookiecutter_keywords = ""
    extended_keywords.extend(k.strip() for k in cookiecutter_keywords.split(","))
    return _json.dumps(extended_keywords)


@_simple_filter  # type: ignore[misc]
def default_codeowners(cookiecutter: dict[str, str]) -> str:
    """Build a default description for the project.

    Args:
        cookiecutter: The cookiecutter context.

    Returns:
        The default description.
    """
    repo_type = cookiecutter["type"]

    codeowners = cookiecutter["default_codeowners"]
    default = _get_from_json("default_codeowners")
    if codeowners != default:
        return codeowners

    github_org = _get_from_json("github_org")
    if github_org != "frequenz-floss":
        return f"TODO(cookiecutter): Add codeowners (like @{github_org}/some-team)"

    type_to_team = {
        "actor": "TODO(cookiecutter): Add codeowners (like @{github_org}/some-team)"
        "# Temporary, should probably change",
        "api": "@frequenz-floss/api-team",
        "lib": "@frequenz-floss/python-sdk-team",
        "app": "@frequenz-floss/python-sdk-team @frequenz-floss/datasci-team "
        "# Temporary, should probably change",
        "model": "@frequenz-floss/datasci-team",
    }

    assert repo_type in type_to_team, f"Unhandled repository type {repo_type!r}"

    return type_to_team[repo_type]



# License: MIT
# Copyright © {% now 'utc', '%Y' %} {{cookiecutter.author_name}}

"""Tests for the {{cookiecutter.name}} package."""

{%- if cookiecutter.type == "api" %}


def test_package_import() -> None:
    """Test that the package can be imported."""
    # pylint: disable=import-outside-toplevel
    from frequenz.api import {{cookiecutter.name}}

    assert {{cookiecutter.name}} is not None


def test_module_import_components() -> None:
    """Test that the modules can be imported."""
    # pylint: disable=import-outside-toplevel
    from frequenz.api.{{cookiecutter.name}} import {{cookiecutter.name}}_pb2

    assert {{cookiecutter.name}}_pb2 is not None

    # pylint: disable=import-outside-toplevel
    from frequenz.api.{{cookiecutter.name}} import {{cookiecutter.name}}_pb2_grpc

    assert {{cookiecutter.name}}_pb2_grpc is not None
{%- else %}
import pytest

from {{cookiecutter.python_package}} import delete_me


def test_{{cookiecutter.name}}_succeeds() -> None:  # TODO(cookiecutter): Remove
    """Test that the delete_me function succeeds."""
    assert delete_me() is True


def test_{{cookiecutter.name}}_fails() -> None:  # TODO(cookiecutter): Remove
    """Test that the delete_me function fails."""
    with pytest.raises(RuntimeError, match="This function should be removed!"):
        delete_me(blow_up=True)
{%- endif %}



# License: MIT
# Copyright © 2023 Frequenz Energy-as-a-Service GmbH

"""Generation tests for cookiecutter."""

import pathlib
import re
import subprocess

import pytest

from frequenz.repo import config


@pytest.mark.integration
@pytest.mark.parametrize("repo_type", [t.value for t in config.RepositoryType])
def test_generation(tmp_path: pathlib.Path, repo_type: str) -> None:
    """Test generation of a new repo."""
    cwd = pathlib.Path().cwd()
    _run(
        tmp_path,
        "cookiecutter",
        "--no-input",
        str(cwd / "cookiecutter"),
        f"type={repo_type}",
        "name=test",
        "description=Test description",
    )

    subdirs = list(tmp_path.iterdir())
    assert len(subdirs) == 1
    repo_path = subdirs[0]
    _run(repo_path, "python3", "-m", "venv", ".venv")

    _update_pyproject_repo_config_dep(
        repo_config_path=cwd, repo_type=repo_type, repo_path=repo_path
    )

    cmd = ". .venv/bin/activate; pip install .[dev-noxfile]; nox -e ci_checks_max pytest_min"
    print()
    print(f"Running in shell [{cwd}]: {cmd}")
    subprocess.run(cmd, shell=True, cwd=repo_path, check=True)


def _run(cwd: pathlib.Path, *cmd: str) -> subprocess.CompletedProcess[bytes]:
    print()
    print("-" * 80)
    print(f"Running [{cwd}]: {' '.join(cmd)}")
    print()
    return subprocess.run(cmd, cwd=cwd, check=True)


def _update_pyproject_repo_config_dep(
    *, repo_config_path: pathlib.Path, repo_type: str, repo_path: pathlib.Path
) -> None:
    """Update the repo config dependency in the generated pyproject.toml.

    This is necessary to make sure we are testing the local version of
    `frequenz-repo-config`, otherwise tests will fail because they will be running with
    an older (released) version of `frequenz-repo-config`.

    Args:
        repo_config_path: Path to the local `frequenz-repo-config` repo.
        repo_type: Type of the repo to generate.
        repo_path: Path to the generated repo.
    """
    repo_config_dep = f"frequenz-repo-config[{repo_type}] @ file://{repo_config_path}"
    repo_config_dep_re = re.compile(rf"""frequenz-repo-config\[{repo_type}\][^"]+""")

    with open(repo_path / "pyproject.toml", encoding="utf8") as pyproject_file:
        pyproject_content = pyproject_file.read()

    with open(repo_path / "pyproject.toml", "w", encoding="utf8") as pyproject_file:
        pyproject_file.write(repo_config_dep_re.sub(repo_config_dep, pyproject_content))



# License: MIT
# Copyright © 2023 Frequenz Energy-as-a-Service GmbH

"""Tests for the core module."""

from frequenz.repo.config import RepositoryType

from . import utils


def test_repository_type_doesnt_have_new_types() -> None:
    """Test that `RepositoryType` doesn't have new types."""
    expected = {
        "actor",
        "api",
        "app",
        "lib",
        "model",
    }
    defined = {t.value for t in RepositoryType}
    assert defined == expected, utils.MSG_UNEXPECTED_REPO_TYPES



# License: MIT
# Copyright © 2023 Frequenz Energy-as-a-Service GmbH

"""Tests for the nox.default module."""

from frequenz.repo import config
from frequenz.repo.config.nox import default

from .. import utils


def test_all_repository_types_are_handled() -> None:
    """Test that all repository types handled match the defined in RepositoryType."""
    expected = {
        *{f"{repo_type.value}_config" for repo_type in config.RepositoryType},
        *{f"{repo_type.value}_command_options" for repo_type in config.RepositoryType},
        "common_config",
        "common_command_options",
    }

    defined = {
        f
        for f in dir(default)
        if not f.startswith("_")
        and (f.endswith("_config") or f.endswith("_command_options"))
    }
    assert defined == expected, utils.MSG_UNEXPECTED_REPO_TYPES



# License: MIT
# Copyright © 2023 Frequenz Energy-as-a-Service GmbH

"""Configuration file for nox."""

from frequenz.repo.config import nox
from frequenz.repo.config.nox import default

config = default.lib_config.copy()
config.extra_paths.extend(
    [
        "cookiecutter/hooks",
        "cookiecutter/local_extensions.py",
    ]
)
nox.configure(config)



# License: MIT
# Copyright © 2023 Frequenz Energy-as-a-Service GmbH

"""Cookiecutter post-generation hooks.

This module contains post-generation hooks to adjust the generated project to the
different types of projects and do some final checks and configuration.
"""

import configparser as _configparser
import dataclasses as _dataclasses
import json as _json
import pathlib as _pathlib
import shutil as _shutil
import subprocess as _subprocess
from collections import namedtuple
from typing import Any


def to_named_tuple(dictionary: dict[Any, Any], /) -> Any:
    """Convert a dictionary to a named tuple.

    Args:
        dictionary: The dictionary to convert.

    Returns:
        The named tuple with the same keys and values as the dictionary.
    """
    filtered = {k: v for k, v in dictionary.items() if not k.startswith("_")}
    return namedtuple("Cookiecutter", filtered.keys())(*filtered.values())


cookiecutter = to_named_tuple(_json.loads("""{{cookiecutter | tojson}}"""))


def main() -> None:
    """Run some post-generation steps.

    This function is called after the project has been generated and it performs
    some fixes for different types of projects and some sanity checks.
    """
    finish_setup()

    print()
    print("-" * 80)
    print()
    success(f"Your 🍪 {cookiecutter.github_repo_name} has been cut!")
    print()
    print_generated_tree()
    print("Here is a list of things that should be reviewed and FIXED:")
    print_todos()
    print()
    print(
        "After completing it you can amend the previous commit using `git commit "
        "--amend` or create a new commit for the changes using `git commit`."
    )
    print()
    print(
        "You can make sure linting and tests pass by creating a virtual "
        "environment, installing the development dependencies and running `nox`:"
    )
    print()
    print(f"cd {cookiecutter.github_repo_name}")
    print("# Requires at least python version 3.11")
    print("python3 -m venv .venv")
    print(". .venv/bin/activate")
    print("pip install .[dev-noxfile]")
    print("nox")
    print()
    print("To generate and serve the documentation:")
    print("pip install .[dev-mkdocs]")
    if cookiecutter.type == "api":
        print("# Requires docker")
    print("mkdocs serve")
    print()
    print("To initialize the GitHub pages website:")
    print("mike deploy --update-aliases next latest")
    print("mike set-default latest")
    print("git push upstream gh-pages  # or origin if you haven't forked the repo")
    print()
    print("Make sure that GitHub pages is enabled in your repository settings:")
    print(
        f"https://github.com/{cookiecutter.github_org}/{cookiecutter.github_repo_name}"
        "/settings/pages"
    )
    print("If all went well, your new website should be available soon at:")
    print(
        f"https://{cookiecutter.github_org}.github.io/{cookiecutter.github_repo_name}/"
    )
    print()
    if warnings := do_sanity_checks():
        for warning in warnings:
            warn(warning)
        print()
    note(
        "If you had any issues or find any errors in the generated files, "
        "please report them!"
    )
    note(
        "https://github.com/frequenz-floss/frequenz-repo-config-python/issues/new/choose"
    )
    print()


def finish_setup() -> None:
    """Finish the setup.

    This function is called after the project has been generated and can be used
    to perform any additional setup steps that are required for the project.
    """
    was_repo_initialized = initialize_git_repo()

    copy_replay_file()

    remove_unneeded_files()

    match cookiecutter.type:
        case "actor":
            finish_actor_setup()
        case "api":
            finish_api_setup()
        case "app":
            finish_app_setup()
        case "lib":
            finish_lib_setup()
        case "model":
            finish_model_setup()

    initialize_git_submodules()
    commit_git_changes(first_commit=was_repo_initialized)


def copy_replay_file() -> None:
    """Copy the replay file to the project root."""
    src = _pathlib.Path("~/.cookiecutter_replay/cookiecutter.json").expanduser()
    dst = _pathlib.Path(".cookiecutter-replay.json")

    if dst.exists():
        print(f"Replay file {dst} already exists. Skipping...")
        return

    if not src.exists():
        print(f"WARNING: No replay file found in {src}. Skipping...")

    try:
        _shutil.copyfile(src, dst)
    except (OSError, IOError) as error:
        print(
            f"WARNING: Error copying the replay file {src} -> {dst} ({error}). "
            "Skipping..."
        )


def initialize_git_submodules() -> bool:
    """Initialize git submodules.

    If a `.gitmodules` file exists and it is not empty, it will be used to initialize
    the git submodules using `git submodule update --init --recursive`.

    If an empty `.gitmodules` file exists, it will be deleted.

    Returns:
        Whether any git submodules were initialized.
    """
    gitmodules_path = _pathlib.Path(".gitmodules")

    if not gitmodules_path.exists():
        return False

    if not gitmodules_path.is_file():
        warn("`.gitmodules` exists but is not a file! Ignoring...")
        return False

    if is_file_empty(gitmodules_path):
        gitmodules_path.unlink()
        return False

    gitmodules_config = _configparser.ConfigParser()
    with gitmodules_path.open("r", encoding="utf8") as file_handle:
        gitmodules_config.read_file(file_handle)

    @_dataclasses.dataclass(frozen=True)
    class Submodule:
        """A git submodule."""

        name: str
        path: str
        url: str

        def mkdir(self) -> None:
            """Create the directory for the submodule."""
            _pathlib.Path(self.path).parent.mkdir(exist_ok=True, parents=True)

    submodules: list[Submodule] = []
    for section in gitmodules_config.sections():
        heading = 'submodule "'
        trailing = '"'
        if not section.startswith(heading) or not section.endswith(trailing):
            warn(f"Can't parse `.gitmodules` section `{section}`! Ignoring...")
            continue
        if "path" not in gitmodules_config[section]:
            warn(
                f"`.gitmodules` contains section `{section}` without `path`! Ignoring..."
            )
            continue
        if "url" not in gitmodules_config[section]:
            warn(
                f"`.gitmodules` contains section `{section}` without `url`! Ignoring..."
            )
            continue
        submodule = Submodule(
            name=section[len(heading) : -len(trailing)],
            path=gitmodules_config[section]["path"],
            url=gitmodules_config[section]["url"],
        )
        # submodule.mkdir()
        submodules.append(submodule)

    if not submodules:
        warn("`.gitmodules` does not contain any valid submodules! Ignoring...")
        return False

    print()
    note("Initializing git submodules...")

    for submodule in submodules:
        try_run(
            [
                "git",
                "submodule",
                "add",
                "--name",
                submodule.name,
                submodule.url,
                submodule.path,
            ],
            verbose=True,
            warn_on_error=True,
            warn_on_bad_status=f"Failed to add submodule `{submodule.name}`!",
            note_on_failure=f"Please add submodule `{submodule.name}` manually.",
        )
    try_run(
        ["git", "submodule", "update", "--init"],
        verbose=True,
        warn_on_error=True,
        warn_on_bad_status="Failed to initialize git submodules!",
        note_on_failure="Please initialize git submodules manually.",
    )
    return True


def initialize_git_repo() -> bool:
    """Initialize a git repository.

    If the project is not a git repository, it will be initialized using
    `git init`.

    If the project is already a git repository, it will be left untouched.

    Returns:
        Whether the project was initialized as a git repository.
    """
    if _pathlib.Path(".git").exists():
        return False
    print()
    note("Initializing git repository...")
    try_run(
        ["git", "init"],
        verbose=True,
        warn_on_error=True,
        warn_on_bad_status="Failed to initialize the git repository!",
        note_on_failure="Please initialize the git repository manually.",
    )
    return True


def commit_git_changes(*, first_commit: bool) -> None:
    """Commit all changes in the git repository.

    If there are no changes to commit, this function does nothing.

    Args:
        first_commit: Whether this is the first commit in the git repository.
    """
    if not try_run(["git", "status", "--porcelain"]):
        return

    print()
    try_run(
        ["git", "add", "."],
        verbose=True,
        warn_on_error=True,
        warn_on_bad_status="Failed to add all changes to the git repository!",
        note_on_failure="Please add all changes to the git repository manually.",
    )
    message = (
        "Initial commit" if first_commit else "Regenerate repository using repo-config"
    )
    try_run(
        ["git", "commit", "-s", "-m", message],
        verbose=True,
        warn_on_error=True,
        warn_on_bad_status="Failed to commit all changes to the git repository!",
        note_on_failure="Please commit all changes to the git repository manually.",
    )


def remove_unneeded_files() -> None:
    """Remove unneeded files.

    Files that are not needed for the project are removed.
    """
    if cookiecutter.license != "MIT":
        _pathlib.Path("LICENSE").unlink()

    if cookiecutter.type != "api":
        _shutil.rmtree("proto")


def is_file_empty(path: _pathlib.Path) -> bool:
    """Check if the file is empty.

    A file is also considered empty if it only contains white spaces.

    Note:
        Only the first 1024 bytes of the file are read for performance and security
        reasons, a files with the first 1024 bytes being white spaces is considered
        empty.

    Args:
        path: The path to the file to check.

    Returns:
        True if the file has only white spaces, False otherwise.
    """
    with path.open("r", encoding="utf8") as file_handle:
        contents = file_handle.read(1024).strip()
        return not contents


def print_generated_tree() -> None:
    """Print the generated files tree."""
    result = try_run(["tree"])
    if result is not None and result.returncode == 0:
        print()


def print_todos() -> None:
    """Print all TODOs in the generated project."""
    todo_str = "TODO(cookiecutter):"
    repo = cookiecutter.github_repo_name
    cmd = ["grep", "-r", "--color", rf"\<{todo_str}.*", "."]
    try_run(
        cmd,
        warn_on_error=True,
        warn_on_bad_status=f"No `{todo_str}` found using `{' '.join(cmd)}`",
        note_on_failure=f"Please search for `{todo_str}` in `{repo}/` manually.",
    )
    print()
    note(
        "Make sure to (create and) configure your GitHub repository too: "
        "https://github.com/frequenz-floss/frequenz-repo-config-python/"
        "wiki/Configuring-a-new-GitHub-repository"
    )


def do_sanity_checks() -> list[str]:
    """Perform sanity checks on the generated project.

    Returns:
        List of warnings.
    """
    warnings: list[str] = []

    if cookiecutter.github_org == "frequenz-floss" and cookiecutter.license != "MIT":
        warnings.append(
            "Using a non-MIT license with a frequenz-floss project is not recommended."
        )

    return warnings


def finish_actor_setup() -> None:
    """Finish the setup for an actor.

    This function is called after the project has been generated and can be used
    to perform any additional setup steps that are required for an actor.
    """


def finish_api_setup() -> None:
    """Finish the setup for an API.

    This function is called after the project has been generated and can be used
    to perform any additional setup steps that are required for an API.

    Operations performed by this function:

    * Rename `src` to `py`
    * Rename `tests` to `pytests`
    """
    recursive_overwrite_move(_pathlib.Path("src"), _pathlib.Path("py"))
    recursive_overwrite_move(_pathlib.Path("tests"), _pathlib.Path("pytests"))


def finish_app_setup() -> None:
    """Finish the setup for an app.

    This function is called after the project has been generated and can be used
    to perform any additional setup steps that are required for an app.
    """


def finish_lib_setup() -> None:
    """Finish the setup for a library.

    This function is called after the project has been generated and can be used
    to perform any additional setup steps that are required for a library.

    Operations performed by this function:

    * Avoid having a subfolder for the type

      - `lib`: `src/frequenz/{name}`
      - `rest`: `src/frequenz/{type}/{name}`
    """
    recursive_overwrite_move(
        _pathlib.Path(f"src/frequenz/{cookiecutter.type}/{cookiecutter.name}"),
        _pathlib.Path(f"src/frequenz/{cookiecutter.name}"),
    )
    _pathlib.Path(f"src/frequenz/{cookiecutter.type}").rmdir()


def finish_model_setup() -> None:
    """Finish the setup for a model.

    This function is called after the project has been generated and can be used
    to perform any additional setup steps that are required for a model.
    """


def try_run(
    cmd: list[str],
    /,
    *,
    warn_on_error: bool = False,
    warn_on_bad_status: str | None = None,
    note_on_failure: str | None = None,
    verbose: bool = False,
) -> _subprocess.CompletedProcess[Any] | None:
    """Try to run a command.

    Args:
        cmd: The command to run.
        warn_on_error: Whether to warn on errors (like command not found) of fail
            silently.
        warn_on_bad_status: If not `None`, warn if the command fails with a non-zero
            status code and use this string as the warning message (the status code
            will be appended to the message).
        note_on_failure: If not `None`, print this note if the command fails (either
            because of an error or a non-zero status code).
        verbose: Whether to print the command before running it.

    Returns:
        The result of the command or `None` if the command could not be run because
        of an error.
    """
    result = None
    failed = False
    if verbose:
        print(f"Executing: {' '.join(cmd)}")
    try:
        result = _subprocess.run(cmd, check=False)
    except (OSError, _subprocess.CalledProcessError) as exc:
        if warn_on_error:
            warn(f"Failed to run the search command `{' '.join(cmd)}`: {exc}")
        failed = True
    else:
        if result.returncode != 0:
            if warn_on_bad_status is not None:
                warn(f"{warn_on_bad_status} (status code: {result.returncode})")
            failed = True
    if failed and note_on_failure is not None:
        note(note_on_failure)

    return result


def recursive_overwrite_move(src: _pathlib.Path, dst: _pathlib.Path) -> None:
    """Recursively move a directory overwriting the target files if they exist.

    Useful when overwriting an existing project to update it.

    Args:
        src: The source directory.
        dst: The target directory.
    """
    _shutil.copytree(src, dst, dirs_exist_ok=True)
    _shutil.rmtree(src)


def success(message: str) -> None:
    """Print a success message.

    Args:
        message: The message to print.
    """
    print(f"\033[32m{message}\033[0m")


def note(message: str) -> None:
    """Print a note.

    Args:
        message: The message to print.
    """
    print(f"\033[36m{message}\033[0m")


def warn(message: str) -> None:
    """Print a warning.

    Args:
        message: The message to print.
    """
    print(f"\033[33mWARNING: {message}\033[0m")


main()



# License: MIT
# Copyright © {% now 'utc', '%Y' %} {{cookiecutter.author_name}}

"""Generate the code reference pages."""

from frequenz.repo.config import mkdocs

mkdocs.generate_python_api_pages("{{cookiecutter | src_path}}", "{{'python-' if cookiecutter.type == 'api'}}reference")
{%- if cookiecutter.type == 'api' %}
mkdocs.generate_protobuf_api_pages()
{%- endif %}
